"""
Tahmid F. Mehdi
Moses Lab, University of Toronto
Helper functions for Variational Inference
July 26, 2018

Copyright 2018 Tahmid Mehdi
This file is part of dphmix.

dphmix is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

dphmix is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with dphmix.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
from .distributions import *
from joblib import Parallel, delayed
import scipy.special as sp


def update_hyperparameters(df, hypers, Ez, k, alpha, e_mu):
    """
    Update the hyperparameters of all distributions with their posterior values

    :param pd.DataFrame df: data
    :param hyperparameters hypers: hyperparameters of distributions
    :param np.ndarray Ez: probability matrix
    :param int>=0 k: cluster index
    :param float>0 alpha: alpha parameter of Dirichlet Process
    :param float e_mu: E(mu)
    :return: updated hyperparameters
    :rtype: hyperparameters
    """

    nu = hypers['nu']
    rho = hypers['rho']
    a = hypers['a']
    b = hypers['b']
    gamma = hypers['gamma']
    delta = hypers['delta']
    eps = hypers['eps']
    zeta = hypers['zeta']
    n_cts = len(nu)
    n_bin = len(gamma)
    n_ord = len(eps)

    post_rho = list(map(lambda col: rho[col]+np.sum(Ez[:, k]), range(n_cts)))
    post_nu = list(map(lambda col: (1/post_rho[col])*(rho[col]*nu[col]+np.sum(Ez[:, k]*df.iloc[:, col].values)),
                       range(n_cts)))
    post_a = list(map(lambda col: a[col]+0.5*np.sum(Ez[:, k]), range(n_cts)))
    post_b = list(map(lambda col: b[col]+0.5*(np.sum(Ez[:, k]*(df.iloc[:, col].values-e_mu[col])**2)
                                              + rho[col]*(e_mu[col]-nu[col])**2), range(n_cts)))
    post_gamma = list(map(lambda col: gamma[col]+np.sum(Ez[:, k]*df.iloc[:, n_cts+col].values), range(n_bin)))
    post_delta = list(map(lambda col: delta[col]+np.sum(Ez[:, k]*(1-df.iloc[:, n_cts+col].values)), range(n_bin)))
    post_eps = list(map(lambda col: eps[col]+np.sum(Ez[:, k]*df.iloc[:, n_cts+n_bin+col].values), range(n_ord)))
    post_zeta = list(map(lambda col: zeta[col]+np.sum(Ez[:, k]), range(n_ord)))
    u1 = 1+np.sum(Ez[:, k])
    u2 = alpha+np.sum(Ez[:, (k+1):].sum(axis=0))
    return dict(nu=post_nu, rho=post_rho, a=post_a, b=post_b,
                gamma=post_gamma, delta=post_delta, eps=post_eps, zeta=post_zeta,
                sb1=u1, sb2=u2)


def update_parameters(hypers):
    """
    Update the parameters of all distributions with their expectations

    :param hyperparameters hypers: hyperparameters of distributions
    :return: moments of parameters
    :rtype: variational expectations
    """

    nu = hypers['nu']
    rho = hypers['rho']
    a = hypers['a']
    b = hypers['b']
    gamma = hypers['gamma']
    delta = hypers['delta']
    eps = hypers['eps']
    zeta = hypers['zeta']
    sb1 = hypers['sb1']
    sb2 = hypers['sb2']
    n_cts = len(nu)
    n_bin = len(gamma)
    n_ord = len(eps)

    e_lntau = np.array(list(map(lambda col: sp.digamma(a[col])-np.log(b[col]), range(n_cts))))
    e_tau = np.array(list(map(lambda col: a[col]/b[col], range(n_cts))))
    e_mu = np.array(list(map(lambda col: nu[col], range(n_cts))))
    e_lnp = np.array(list(map(lambda col: sp.digamma(gamma[col])-sp.digamma(gamma[col]+delta[col]), range(n_bin))))
    e_lnip = np.array(list(map(lambda col: sp.digamma(delta[col])-sp.digamma(gamma[col]+delta[col]), range(n_bin))))
    e_lnlam = np.array(list(map(lambda col: sp.digamma(eps[col])-np.log(zeta[col]), range(n_ord))))
    e_lam = np.array(list(map(lambda col: eps[col]/zeta[col], range(n_ord))))
    e_lnv = sp.digamma(sb1)-sp.digamma(sb1+sb2)
    e_lniv = sp.digamma(sb2)-sp.digamma(sb1+sb2)

    return dict(lntau=e_lntau,tau=e_tau,mu=e_mu, lnp=e_lnp,lnip=e_lnip, lnlam=e_lnlam,lam=e_lam, lnv=e_lnv,lniv=e_lniv)


def calc_expectations(x, lnFactorial, pars):
    """
    Helps update_expectations

    :param pd.Series x: data
    :param list lnFactorial: each element is the ln of the factorial of a poisson variate
    :param variational expectations pars: parameters for updating probability matrix
    :return: an element for the probability matrix before exponentiation
    :rtype: float
    """

    e_lntau = pars['lntau']
    e_tau = pars['tau']
    e_mu = pars['mu']
    e_lnp = pars['lnp']
    e_lnip = pars['lnip']
    e_lnlam = pars['lnlam']
    e_lam = pars['lam']

    n_cts = len(e_mu)
    n_bin = len(e_lnp)

    e = 0.5*np.sum(e_lntau-e_tau*(x[:n_cts]-e_mu)**2) \
        + np.sum(x[n_cts:(n_cts+n_bin)]*e_lnp+(1-x[n_cts:(n_cts+n_bin)])*e_lnip) \
        + np.sum(x[(n_cts+n_bin):]*e_lnlam-lnFactorial-e_lam)

    return e


def update_expectation(x, lnFactorial, pars):
    """
    Update expectation matrix

    :param pd.Series x: data
    :param list lnFactorial: each element is the ln of the factorial of a poisson variate
    :param list pars: each element is a variational expectation
    :return: a column of the probability matrix
    :rtype: list
    """

    max_clusters = len(pars)
    en = list(map(lambda t: calc_expectations(x, lnFactorial, pars[t]), range(max_clusters)))
    return en


def elbo_loglik_cluster(x, pars):
    """
    Calculate E[lnP(x|z,pars)] for a cluster with parameters pars

    :param pd.Series x: data
    :param variational expectations pars: parameters for calculating ELBO
    :return: part of the ELBO
    :rtype: float
    """

    e_lntau = pars['lntau']
    e_tau = pars['tau']
    e_mu = pars['mu']
    e_lnp = pars['lnp']
    e_lnip = pars['lnip']
    e_lnlam = pars['lnlam']
    e_lam = pars['lam']
    n_cts = len(e_mu)
    n_bin = len(e_lnp)
    n_ord = len(e_lam)
    sum_ln_normal = np.sum(list(map(lambda G: ln_normal(x[G], e_lntau[G], e_mu[G], e_tau[G]), range(n_cts))))
    sum_ln_bernoulli = np.sum(list(map(lambda B: ln_bernoulli(x[n_cts+B], e_lnp[B], e_lnip[B]), range(n_bin))))
    sum_ln_poisson = np.sum(list(map(lambda P: ln_poisson(x[n_cts+n_bin+P], e_lnlam[P], e_lam[P]), range(n_ord))))
    return sum_ln_normal+sum_ln_bernoulli+sum_ln_poisson


def elbo_loglik(x, pars):
    """
    Calculate E[lnP(x|z,pars)]

    :param pd.Series x: data
    :param list pars: elements are variational expectations
    :return: E[lnP(x|z,pars)] for ELBO
    :rtype: float
    """

    max_clusters = len(pars)
    sum = list(map(lambda k: elbo_loglik_cluster(x, pars[k]), range(max_clusters)))
    return sum


def elbo_prior(pars, hypers):
    """
    Calculate E[lnP(pars|hypers)]

    :param variational expectations pars: parameters for ELBO
    :param hyperparameters hypers: hyperparameters for distributions
    :return: E[lnP(pars|hypers)]
    :rtype: float
    """

    nu = hypers['nu']
    rho = hypers['rho']
    a = hypers['a']
    b = hypers['b']
    gamma = hypers['gamma']
    delta = hypers['delta']
    eps = hypers['eps']
    zeta = hypers['zeta']

    e_lntau = pars['lntau']
    e_tau = pars['tau']
    e_mu = pars['mu']
    e_lnp = pars['lnp']
    e_lnip = pars['lnip']
    e_lnlam = pars['lnlam']
    e_lam = pars['lam']
    n_cts = len(e_mu)
    n_bin = len(e_lnp)
    n_ord = len(e_lam)
    sum_ln_ng = np.sum(list(map(lambda G: ln_normalgamma(e_mu[G], e_tau[G], e_lntau[G], nu[G],rho[G], a[G],b[G]),
                                range(n_cts))))
    sum_ln_beta = np.sum(list(map(lambda B: ln_beta(e_lnp[B], e_lnip[B], gamma[B], delta[B]), range(n_bin))))
    sum_ln_gamma = np.sum(list(map(lambda P: ln_gamma(e_lam[P], e_lnlam[P], eps[P], zeta[P]), range(n_ord))))
    return sum_ln_ng+sum_ln_beta+sum_ln_gamma


def local_sum_par(pars, var_hypers):
    """
    Helps parallelize ELBO computation

    :param variational expectations pars: parameters for ELBO
    :param hyperparameters var_hypers: hyperparameters of distributions
    :return: part of ELBO
    :rtype: float
    """

    return elbo_prior(pars, var_hypers)+ln_beta(pars['lnv'], pars['lniv'], var_hypers['sb1'], var_hypers['sb2'])


def elbo(df, pars, var_hypers, prior_hypers, e_z, c_mat, lnPc, alpha, cores):
    """
    Calculate ELBO

    :param pd.DataFrame df: data
    :param variational expectations pars: parameters of ELBO
    :param hyperparameters var_hypers: variational parameters
    :param hyperparameters prior_hypers: prior hyperparameters of distributions
    :param np.ndarray e_z: probability matrix
    :param np.ndarray c_mat: cluster probability matrix
    :param list lnPc: list of lnP(c=[1..T]|v)
    :param float>0 alpha: parameter for Dirichlet Process
    :param int>=1 cores: number of cores to use
    :return: ELBO
    :rtype: float<=0
    """

    n = len(df)
    max_clusters = e_z.shape[1]
    loglik_sums = np.array(Parallel(n_jobs=cores)(delayed(elbo_loglik)(x=df.iloc[idx], pars=pars) for idx in range(n)))
    loglik = np.sum(e_z*loglik_sums)
    logprior = np.sum(Parallel(n_jobs=cores)(delayed(elbo_prior)(pars=pars[k], hypers=prior_hypers)
                                             for k in range(max_clusters)))
    c_mat_colSums = np.sum(c_mat, axis=0)
    lnpz = np.sum(c_mat_colSums*lnPc)
    lnpv = np.sum(Parallel(n_jobs=cores)(delayed(ln_beta)(ln_x=pars[k]['lnv'], ln_ix=pars[k]['lniv'], a=1, b=alpha)
                                         for k in range(max_clusters)))
    local_sum = np.sum(Parallel(n_jobs=cores)(delayed(local_sum_par)(pars=pars[k], var_hypers=var_hypers[k])
                                              for k in range(max_clusters)))
    lnq = np.sum(np.log(np.power(c_mat, c_mat)))+local_sum
    return loglik+logprior+lnpz+lnpv-lnq



"""
Tahmid F. Mehdi
Moses Lab, University of Toronto
Cluster Classes
April 22, 2018

Copyright 2018 Tahmid Mehdi
This file is part of dphmix.

dphmix is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

dphmix is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with dphmix.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np


class GibbsCluster:
    def __init__(self, cluster_assignments, parameters):
        """
        Construct a GibbsCluster object

        :param array-like cluster_assignments: list of clusters
        :param list parameters: parameters of each cluster
        """

        self.c = np.array(cluster_assignments)
        self.phi = parameters
        self.n_clusters = len(np.unique(cluster_assignments))

    def __repr__(self):
        return "{} clusters from a Gibbs Sampler".format(self.n_clusters)

    def __str__(self):
        return "{} clusters from a Gibbs Sampler".format(self.n_clusters)


class MFVICluster:
    def __init__(self, cluster_assignments, var_pars, moments, prob_matrix, cluster_mapping, elbo):
        """
        Construct a MFVICluster object

        :param array-like cluster_assignments: list of clusters
        :param list var_pars: variational parameters
        :param list moments: expectations for cluster parameters
        :param np.ndarray prob_matrix: matrix where the element in row i, column j is the probability of observation i
        being in cluster j
        :param dict cluster_mapping: mapping of cluster indices in the model (0-max_clusters) to indices in
        cluster_assignments
        :param float<=0 elbo: the final ELBO
        """

        self.c = np.array(cluster_assignments)
        self.variational_parameters = var_pars
        self.moments = moments
        self.Ez = prob_matrix
        self.cluster_map = cluster_mapping
        self.n_clusters = len(np.unique(cluster_assignments))
        self.ELBO = elbo

    def __repr__(self):
        return "{} clusters from coordinate ascent mean-field variational inference with ELBO={}"\
            .format(self.n_clusters, self.ELBO)

    def __str__(self):
        return "{} clusters from coordinate ascent mean-field variational inference with ELBO={}" \
            .format(self.n_clusters, self.ELBO)


"""
Tahmid F. Mehdi
Moses Lab, University of Toronto
Unsupervised and Constrained Dirichlet Heterogeneous Mixtures of Exponential Families with Variational Inference
August 9, 2018

References:
DM Blei & MI Jordan. (2006). Variational Inference for Dirichlet Process Mixtures. Bayesian Analysis, 1(1), 121-144.
A Vlachos, A Korhonen & Z Ghahramani. (2009). Unsupervised and Constrained Dirichlet Process Mixture Models for Verb
Clustering. Proceedings of the EACL 2009 Workshop on GEMS: GEometical Models of Natural Language Semantics, 74-82.

Copyright 2018 Tahmid Mehdi
This file is part of dphmix.

dphmix is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

dphmix is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with dphmix.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
from .mfviHelpers import *
from .clusters import *
from joblib import Parallel, delayed
import pandas as pd
from .utils import *


class VariationalDPHM:
    """
Implements Unsupervised & Semi-supervised Mean-field Variational Inference (MFVI) with conjugate priors for Dirichlet
Process Mixture Models where the independent variates follow Normal, Bernoulli & Poisson distributions.

If x has n_cts, n_bin & n_ord continuous, binary & ordinal variates, respectively, then
hyperparameters are contained in dictionaries where keys=['nu','rho','a','b','gamma','delta','eps','zeta'] where
the values of 'nu','rho','a' & 'b' must be lists of numbers with length n_cts
the values of 'gamma' & 'delta' must be lists of numbers with length n_bin
the values of 'eps' & 'zeta' must be lists of numbers with length n_ord

Parameters are contained in dictionaries where keys=['mu','tau','p','lam'] where
the values of 'mu' & 'tau' must be lists of numbers with length n_cts
the value of 'p' must be a list of numbers with length n_bin
the value of 'lam' must be a list of numbers with length n_ord

x | parameters ~ N(mu[1],1/tau[1])..N(mu[n_cts],1/tau[n_cts])Bern(p[1])..Bern(p[n_bin])Poi(lam[1])..Poi(lam[n_ord])
where (mu[i],tau[i]) ~ NormalGamma(nu[i],rho[i],a[i],b[i]), i=1..n_cts
      p[i] ~ Beta(gamma[i],delta[i]), i=1..n_bin
      lam[i] ~ Gamma(eps[i],zeta[i]), i=1..n_ord

For MFVI, stick-break points v[i] ~ Beta(1,alpha) yield mixing weights pi[i]=v[i]*sum(j=1..i-1)(1-v[j]) for
i=1..max_clusters.
z ~ Mult(1, pi)
Add 'sb1' & 'sb2' to hyperparameters for variational posterior of v
Variational expectations are contained in dictionaries where
keys=['lntau','tau','mu','lnp','lnip','lnlam','lam','lnv','lniv'] where
the values of 'mu','tau' & 'lntau' must be lists of numbers with length n_cts
the value of 'lnp' & 'lnip' must be a list of numbers with length n_bin
the value of 'lam' & 'lnlam' must be a list of numbers with length n_ord
the values of 'lnv' & 'lniv' must be numbers
    """
    def __init__(self, alpha, iterations, max_clusters, tol=1e-3, n_jobs=1, random_state=None):
        """
        Constructs a DirichletHeterogeneousMixture object

        :param float>0 alpha: the alpha parameter of the Dirichlet Process distribution
        :param int>=1 iterations: maximum number of iterations
        :param int>=2 max_clusters: maximum number of clusters
        :param float>0 tol: if inference='mfvi', the algorithm stops when the difference between evidence lower bounds
        (ELBOs) in 2 consecutive iterations is < tol
        :param int>=1 n_jobs: number of cores to use
        :param random_state: the random seed for initializing clusters in MFVI
        :type random_state: int>=0 or None
        """

        assert alpha > 0, 'alpha must be positive'
        assert iterations >= 1 and isinstance(iterations, int), 'iterations must be a positive integer'
        assert max_clusters >= 2 and isinstance(max_clusters, int), 'max_clusters must be an integer at least 2'
        assert tol > 0, 'tol must be positive'
        assert n_jobs >= 1 and isinstance(n_jobs, int), 'n_jobs must be an integer at least 1'
        self.alpha = alpha
        self.iterations = iterations
        self.max_clusters = max_clusters
        self.tol = tol
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.var_moments = None

    def get_params(self):
        """
        Returns a dictionary of arguments for the object

        :return: arguments for object
        :rtype: dict
        """
        return dict(alpha=self.alpha, iterations=self.iterations, max_clusters=self.max_clusters, tol=self.tol,
                    n_jobs=self.n_jobs, random_state=self.random_state)

    def __repr__(self):
        return "VariationalDPHM(alpha={}, iterations={}, max_clusters={}, tol={}, n_jobs={}, random_state={})"\
            .format(self.alpha, self.iterations, self.max_clusters, self.tol, self.n_jobs, self.random_state)

    def __str__(self):
        return "VariationalDPHM(alpha={}, iterations={}, max_clusters={}, tol={}, n_jobs={}, random_state={})"\
            .format(self.alpha, self.iterations, self.max_clusters, self.tol, self.n_jobs, self.random_state)

    def _mfvi(self, X, hyperparameters, ml=[]):
        """
        Clusters X with variational inference given must-link constraints ml

        :param pd.DataFrame X: data
        :param hyperparameters hyperparameters: hyperparameters for distributions
        :param list ml: each element is a list of indices of points that must be in the same cluster
        :return: cluster assignment for each observation, variational parameters and moments, and final ELBO
        :rtype: MFVICluster
        """

        n = len(X)
        n_cts = len(hyperparameters['nu'])
        n_bin = len(hyperparameters['gamma'])
        ml_flat = [idx for link in ml for idx in link]
        np.random.seed(self.random_state)  # set random seed
        # initialize expectation matrix. Ez[i,j] is the probability observation i is in cluster j
        Ez = np.zeros((n, self.max_clusters))
        Ez[np.arange(n), np.random.randint(0, self.max_clusters, n)] = 1
        # stick-break points ~ Beta(1,alpha)
        hyperparameters['sb1'] = 1
        hyperparameters['sb2'] = self.alpha
        # initialize hyperparameters & parameters for each cluster
        phi_hyper = [hyperparameters]*self.max_clusters
        phi_par = [update_parameters(hyperparameters)]*self.max_clusters
        # calculate factorials for Poisson variates
        lnFactorials = X.apply(lambda row: [math.log(math.factorial(xi)) for xi in row[(n_cts+n_bin):]], axis=1)
        ELBO = float('-inf')  # initialize ELBO
        for i in range(self.iterations):
            print("Running iteration %s ... " % str(i + 1), end="")
            prevELBO = ELBO
            # update hyperparameters
            phi_hyper = Parallel(n_jobs=self.n_jobs)(delayed(update_hyperparameters)
                                                     (df=X, hypers=hyperparameters, Ez=Ez, k=k, alpha=self.alpha,
                                                      e_mu=phi_par[k]['mu']) for k in range(self.max_clusters))
            # update parameters
            phi_par = Parallel(n_jobs=self.n_jobs)(delayed(update_parameters)(hypers=phi_hyper[k])
                                                   for k in range(self.max_clusters))
            # update E(z_nk) for all n,k
            Ez = Parallel(n_jobs=self.n_jobs)(delayed(update_expectation)(x=X.iloc[idx], lnFactorial=lnFactorials[idx],
                                                                          pars=phi_par) for idx in range(n))
            Ez = np.array(Ez)
            # calculate joint probabilities for must-link data
            for link in ml:
                Ez[link, :] = np.sum(Ez[link, :], axis=0)
            # extract E[ln(1-v)] for each cluster and lnP(c=t|v) for each t=1..max_clusters
            lnivs = [phi_par[t]['lniv'] for t in range(self.max_clusters)]
            lnPc = [phi_par[t]['lnv']+np.sum(lnivs[:t]) for t in range(self.max_clusters)]
            # Add row vector lnP(c=[1..max_clusters]|v) to each row of Ez
            Ez = Ez + lnPc
            # the exp-normalize trick for preventing underflow
            rowMax = np.max(Ez, axis=1).reshape((n, 1))
            Ez = np.exp(Ez - rowMax)
            # normalize rows of Ez
            rowSums = Ez.sum(axis=1)[:, None]  # sum probabilities for each observation
            # if all probabilities for an observation are below machine epsilon then it won't be assigned a cluster
            assert 0 not in rowSums, \
                str(list(rowSums).count(0))+' observations could not be assigned to a cluster. Increase max_clusters'
            Ez = Ez/rowSums
            # create a matrix of cluster probabilities without duplicated rows for must-link constraints
            c_mat = np.delete(Ez, ml_flat, axis=0)
            for link in ml:
                c_mat = np.append(c_mat, [Ez[link[0], :]], axis=0)
            # calculate ELBO & gain
            ELBO = elbo(df=X, pars=phi_par,var_hypers=phi_hyper,prior_hypers=hyperparameters, e_z=Ez,
                        c_mat=c_mat, lnPc=lnPc, alpha=self.alpha, cores=self.n_jobs)
            ELBOgain = ELBO - prevELBO
            print("ELBO: %s ... gained %s" % (str(ELBO), str(ELBOgain)))
            if ELBOgain < self.tol:
                print("ELBO converged!")
                break
        # assign to each observation, the cluster it's most likely to belong to from the expectation matrix
        c = np.array(Ez.argmax(axis=1)).reshape(n)
        # map cluster indices so they're all consecutive integers
        uniq_c = np.unique(c)
        # filter hyperparameters & expectations
        var_hyper = [phi_hyper[j] for j in uniq_c]
        self.var_moments = [phi_par[j] for j in uniq_c]
        # a mapping to help read the E(z_nk) matrix
        cluster_map = {}
        for clust in uniq_c:
            cluster_map[clust] = np.where(uniq_c == clust)[0][0]

        clusters = list(map(lambda clust_idx: cluster_map[clust_idx], c))
        solution = MFVICluster(clusters, var_hyper, self.var_moments, Ez, cluster_map, ELBO)
        if solution.n_clusters == self.max_clusters:
            print("Warning: max_clusters reached, you may need to cluster again with a higher max_clusters.")
        print("Done")
        return solution

    def fit_predict(self, X, ml=[], hyperparameters=None):
        """
        Fit the model to X & predict clusters for each observation (row) in X given must-link constraints ml

        :param pd.DataFrame X: data
        :param list ml: each element is a list of indices of points that must be in the same cluster
        :param hyperparameters: hyperparameters of distributions for the variates
        :type hyperparameters: hyperparameters or None
        :return: the clustering solution
        :rtype: MFVICluster
        """

        assert isinstance(X, pd.DataFrame), 'X must be a pandas.DataFrame'
        ml_flat = [idx for link in ml for idx in link]
        assert not(is_duplicate(ml_flat)), 'Integers in ml are not unique'
        assert (0 <= np.array(ml_flat)).all() and (np.array(ml_flat) < len(X)).all() \
            and all([isinstance(item, int) for item in ml_flat]), 'Elements of ml must be integers in [0,len(X))'
        # sort features so continuous variables come first, then binary, followed by non-negative discrete
        X, n_cts, n_bin, n_ord = sort_features(X)
        if hyperparameters is None:
            # set prior hyperparameters
            nu = np.mean(X[X.columns[:n_cts]])
            rho = [1]*n_cts
            a = [1]*n_cts
            b = (np.var(X[X.columns[:n_cts]]))
            gamma = [2]*n_bin
            delta = (2-2*np.mean(X[X.columns[n_cts:(n_cts+n_bin)]]))/np.mean(X[X.columns[n_cts:(n_cts+n_bin)]])
            eps = [1]*n_ord
            zeta = 1/np.mean(X[X.columns[(n_cts+n_bin):]])
            hyperparameters = dict(nu=nu, rho=rho, a=a, b=b, gamma=gamma, delta=delta, eps=eps, zeta=zeta)
        else:  # check lengths of hyperparameters
            assert isinstance(hyperparameters, dict), 'hyperparameters must be a dictionary'
            assert 'nu' in hyperparameters and 'rho' in hyperparameters and 'a' in hyperparameters \
                   and 'b' in hyperparameters and 'gamma' in hyperparameters and 'delta' in hyperparameters \
                   and 'eps' in hyperparameters and 'zeta' in hyperparameters, \
                'hyperparameters needs keys "nu", "rho", "a", "b", "gamma", "delta", "eps" & "zeta"'
            assert len(hyperparameters['nu']) == n_cts and len(hyperparameters['rho']) == n_cts \
                   and len(hyperparameters['a']) == n_cts and len(hyperparameters['b']) == n_cts, \
                'Number of hyperparameters for continuous variates do not match the number of continuous variates'
            assert len(hyperparameters['gamma']) == n_bin and len(hyperparameters['delta']) == n_bin, \
                'Number of hyperparameters for binary variates do not match the number of binary variates'
            assert len(hyperparameters['eps']) == n_ord and len(hyperparameters['zeta']) == n_ord, \
                'Number of hyperparameters for ordinal variates do not match the number of ordinal variates'

        # perform MFVI
        clust_soln = self._mfvi(X, hyperparameters, ml)

        return clust_soln

    def predict(self, X):
        """
        Predict the cluster of each observation (row) in X

        :param pd.DataFrame X: data
        :return: the clustering solution
        :rtype: np.array, np.ndarray
        """

        assert self.var_moments is not None, 'Model has not been fit'
        # sort features so continuous variables come first, then binary, followed by non-negative discrete
        X, n_cts_data, n_bin_data, n_ord_data = sort_features(X)
        n = len(X)
        n_clusters = len(self.var_moments)
        n_cts = len(self.var_moments[0]['mu'])
        n_bin = len(self.var_moments[0]['lnp'])
        n_ord = len(self.var_moments[0]['lam'])
        assert n_cts+n_bin+n_ord == len(X.columns), 'The features in X do not match the features of the model'
        # log factorials of Poisson features
        lnFactorials = X.apply(lambda row: [math.log(math.factorial(xi)) for xi in row[(n_cts+n_bin):]], axis=1)
        # update E(z_nk) for all n,k
        Ez = Parallel(n_jobs=self.n_jobs)(delayed(update_expectation)(x=X.iloc[idx], lnFactorial=lnFactorials[idx],
                                                                      pars=self.var_moments) for idx in range(n))
        Ez = np.array(Ez)
        # extract E[ln(1-v)] for each cluster and lnP(c=t|v) for each t=1..max_clusters
        lnivs = [self.var_moments[t]['lniv'] for t in range(n_clusters)]
        lnPc = [self.var_moments[t]['lnv']+np.sum(lnivs[:t]) for t in range(n_clusters)]
        # Add row vector lnP(c=[1..max_clusters]|v) to each row of Ez
        Ez = Ez + lnPc
        # the exp-normalize trick for preventing underflow
        rowMax = np.max(Ez, axis=1).reshape((n, 1))
        Ez = np.exp(Ez - rowMax)
        # normalize rows of Ez
        rowSums = Ez.sum(axis=1)[:, None]  # sum probabilities for each observation
        # if all probabilities for an observation are below machine epsilon then it won't be assigned a cluster
        assert 0 not in rowSums, \
            str(list(rowSums).count(0)) + ' observations could not be assigned to a cluster. Increase max_clusters'
        Ez = Ez/rowSums
        # pick most probable cluster for each observation
        c = np.array(Ez.argmax(axis=1)).reshape(n)
        return c, Ez



"""
Tahmid F. Mehdi
Moses Lab, University of Toronto
Common functions for GibbsDPHM & VariationalDPHM
July 19, 2018


Copyright 2018 Tahmid Mehdi
This file is part of dphmix.

dphmix is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

dphmix is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with dphmix.  If not, see <http://www.gnu.org/licenses/>.
"""


def is_duplicate(l):
    """
    Determines if l has any duplicate entries

    :param list l: list to check
    :return: whether l has any duplicate entries
    :rtype: bool
    """

    for i in range(len(l)):
        for j in range(i+1, len(l)):
            if l[i] == l[j]:
                return True
    return False


def sort_features(X):
    """
    Order the columns of X so Gaussian variates come first, then Bernoullis then Poissons

    :param pd.DataFrame X: data
    :return: X with sorted features & the numbers of continuous, binary & non-negative discrete features
    :rtype: pd.DataFrame, int[>=0], int[>=0], int[>=0]
    """

    n_cts = 0
    n_bin = 0
    n_ord = 0
    ordered_cols = []
    # make a dictionary where keys are data types & values are column names whose data have that type
    dtypes = X.columns.to_series().groupby(X.dtypes).groups
    datatypes = {k.name: v for k, v in dtypes.items()}
    # store continuous variates first
    if 'float64' in datatypes:
        n_cts = len(datatypes['float64'])
        ordered_cols.extend(datatypes['float64'])

    if 'int64' in datatypes:
        bernoulli = []
        poisson = []
        for feature in datatypes['int64']:  # iterate through integer columns
            # if all values of data[feature] are binary, store it as a Bernoulli
            binaries = [datum for datum in X[feature] if datum == 0 or datum == 1]
            if len(binaries) == len(X):
                n_bin += 1
                bernoulli.append(feature)
            else:
                # if all values of data[feature] are discrete & non-negative, store it as a Poisson
                ordinals = [datum for datum in X[feature] if datum >= 0]
                if len(ordinals) == len(X):
                    n_ord += 1
                    poisson.append(feature)

        ordered_cols.extend(bernoulli)
        ordered_cols.extend(poisson)

    X = X[ordered_cols]
    X.columns = [str(col) for col in X.columns]
    X.index = [str(i) for i in X.index]
    return X, n_cts, n_bin, n_ord



"""
Tahmid F. Mehdi
Moses Lab, University of Toronto
Helper functions for Gibbs Sampler
January 31, 2018

Copyright 2018 Tahmid Mehdi
This file is part of dphmix.

dphmix is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

dphmix is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with dphmix.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
import pandas as pd
from .distributions import *


def sample_parameters(hypers, n_samples):
    """
    Produces a list of n_samples parameters

    :param hyperparameters hypers: hyperparameters for distributions
    :param int>=1 n_samples: number of parameters to generate
    :return: n_samples parameters
    :rtype: list
    """

    nu = hypers['nu']
    rho = hypers['rho']
    a = hypers['a']
    b = hypers['b']
    gamma = hypers['gamma']
    delta = hypers['delta']
    eps = hypers['eps']
    zeta = hypers['zeta']
    n_cts = len(nu)
    n_bin = len(gamma)
    n_ord = len(eps)
    tau = list(map(lambda col: np.random.gamma(a[col], 1/b[col], n_samples), range(n_cts)))
    mu = list(map(lambda col: np.random.normal(nu[col], 1/np.sqrt(rho[col]*tau[col]), n_samples), range(n_cts)))
    p = list(map(lambda col: np.random.beta(gamma[col], delta[col], n_samples), range(n_bin)))
    lam = list(map(lambda col: np.random.gamma(eps[col], 1/zeta[col], n_samples), range(n_ord)))

    par_dicts = list(map(lambda sample: dict(mu=[mi[sample] for mi in mu], tau=[ti[sample] for ti in tau],
                                             p=[pi[sample] for pi in p],
                                             lam=[li[sample] for li in lam]), range(n_samples)))
    return par_dicts


def posterior_hyperparameters(data, hypers):
    """
    Produces posterior hyperparameters given data

    :param data: the data
    :type data: pd.DataFrame or pd.Series
    :param hyperparameters hypers: hyperparameters of distributions
    :return: posterior hyperparameters
    :rtype: hyperparameters
    """

    nu = hypers['nu']
    rho = hypers['rho']
    a = hypers['a']
    b = hypers['b']
    gamma = hypers['gamma']
    delta = hypers['delta']
    eps = hypers['eps']
    zeta = hypers['zeta']
    n_cts = len(nu)
    n_bin = len(gamma)
    n_ord = len(eps)
    if isinstance(data, pd.DataFrame):
        m = len(data)
        x_bar = data.mean(axis=0)

        post_nu = list(map(lambda col: (rho[col]*nu[col]+m*x_bar[col])/(rho[col]+m), range(n_cts)))
        post_rho = list(map(lambda col: rho[col]+m, range(n_cts)))
        post_a = list(map(lambda col: a[col]+m/2, range(n_cts)))
        post_b = list(map(lambda col: b[col]+0.5*np.sum((data.iloc[:, col]-x_bar[col])**2)
                                      + (m*rho[col]*(x_bar[col]-nu[col])**2)/(2*(rho[col]+m)), range(n_cts)))

        post_gamma = list(map(lambda col: gamma[col]+np.sum(data.iloc[:, n_cts+col]), range(n_bin)))
        post_delta = list(map(lambda col: delta[col]+m-np.sum(data.iloc[:, n_cts+col]), range(n_bin)))

        post_eps = list(map(lambda col: eps[col]+np.sum(data.iloc[:, n_cts+n_bin+col]), range(n_ord)))
        post_zeta = list(map(lambda col: zeta[col]+m, range(n_ord)))
    else:
        m = 1
        x_bar = data

        post_nu = list(map(lambda col: (rho[col]*nu[col]+m*x_bar[col])/(rho[col]+m), range(n_cts)))
        post_rho = list(map(lambda col: rho[col]+m, range(n_cts)))
        post_a = list(map(lambda col: a[col]+m/2, range(n_cts)))
        post_b = list(map(lambda col: b[col]+0.5*(data[col]-x_bar[col])**2
                                      + (m*rho[col]*(x_bar[col]-nu[col])**2)/(2*(rho[col]+m)), range(n_cts)))

        post_gamma = list(map(lambda col: gamma[col]+data[n_cts+col], range(n_bin)))
        post_delta = list(map(lambda col: delta[col]+m-data[n_cts+col], range(n_bin)))

        post_eps = list(map(lambda col: eps[col]+data[n_cts+n_bin+col], range(n_ord)))
        post_zeta = list(map(lambda col: zeta[col]+m, range(n_ord)))

    return dict(nu=post_nu, rho=post_rho, a=post_a, b=post_b,
                gamma=post_gamma, delta=post_delta, eps=post_eps, zeta=post_zeta)


def likelihood(data, theta):
    """
    Calculates likelihood

    :param data: the data
    :type data: pd.DataFrame or pd.Series
    :param parameters theta: parameters of the distributions
    :return: P(data|theta)
    :rtype: 0<=float<=1
    """

    mu = theta['mu']
    tau = theta['tau']
    p = theta['p']
    lam = theta['lam']
    n_cts = len(mu)
    n_bin = len(p)
    n_ord = len(lam)
    product = 1
    if isinstance(data, pd.DataFrame):
        for i in range(len(data)):
            for G in range(n_cts):
                product *= normal_pdf(data.iloc[i, G], mu[G], tau[G])

            for B in range(n_bin):
                product *= bernoulli_pmf(data.iloc[i, n_cts+B], p[B])

            for P in range(n_ord):
                product *= poisson_pmf(data.iloc[i, n_cts+n_bin+P], lam[P])

    else:
        for G in range(n_cts):
            product *= normal_pdf(data[G], mu[G], tau[G])

        for B in range(n_bin):
            product *= bernoulli_pmf(data[n_cts+B], p[B])

        for P in range(n_ord):
            product *= poisson_pmf(data[n_cts+n_bin+P], lam[P])

    return product


def assign_cluster(prob_dict):
    """
    Assign a cluster based on probabilities in prob_dict

    :param dict prob_dict: mapping of clusters to probabilities
    :return: a cluster
    :rtype: int
    """

    rand = np.random.uniform(0, sum(prob_dict.values()))
    total = 0
    for clust_idx, prob in prob_dict.items():
        total += prob
        if rand <= total:
            return clust_idx
    assert False, 'Error: No Cluster Assigned'


def integrate(data, hypers):
    """
    Compute integral of P(x|parameters)P(parameters) over parameter space

    :param data: the data
    :type data: pd.DataFrame or pd.Series
    :param hyperparameters hypers: hyperparameters of distributions
    :return: integral of P(x|parameters)P(parameters) over parameter space
    :rtype: 0<=float<=1
    """

    nu = hypers['nu']
    rho = hypers['rho']
    a = hypers['a']
    b = hypers['b']
    gamma = hypers['gamma']
    delta = hypers['delta']
    eps = hypers['eps']
    zeta = hypers['zeta']
    post_hypers = posterior_hyperparameters(data, hypers)
    post_nu = post_hypers['nu']
    post_rho = post_hypers['rho']
    post_a = post_hypers['a']
    post_b = post_hypers['b']
    post_gamma = post_hypers['gamma']
    post_delta = post_hypers['delta']
    post_eps = post_hypers['eps']
    post_zeta = post_hypers['zeta']
    n_cts = len(rho)
    n_bin = len(gamma)
    n_ord = len(eps)
    product = 1
    for G in range(n_cts):
        product *= np.sqrt(rho[G]/post_rho[G])*(b[G]**a[G])/(post_b[G]**post_a[G])*sp.gamma(post_a[G])/sp.gamma(a[G])

    for B in range(n_bin):
        product *= sp.beta(post_gamma[B], post_delta[B])/sp.beta(gamma[B], delta[B])

    if isinstance(data, pd.DataFrame):
        product *= (2*np.pi)**(-n_cts*len(data)/2)
        for P in range(n_ord):
            product *= (zeta[P]**eps[P])*sp.gamma(post_eps[P])/((post_zeta[P]**post_eps[P])*sp.gamma(eps[P]))
            for i in range(len(data)):
                product *= 1/math.factorial(data.iloc[i, n_cts+n_bin+P])

    else:
        product *= (2*np.pi)**(-n_cts/2)
        for P in range(n_ord):
            product *= (zeta[P]**eps[P])*sp.gamma(post_eps[P])\
                       / ((post_zeta[P]**post_eps[P])*sp.gamma(eps[P])*math.factorial(data[n_cts+n_bin+P]))

    return product


def sample_parameters_posterior(data, hypers):
    """
    Sample parameters from posterior distribution

    :param pd.DataFrame data: data
    :param hyperparameters hypers: hyperparameters of distributions
    :return: parameters from posterior
    :rtype: parameters
    """

    if len(data) > 0:
        post_hyperparameters = posterior_hyperparameters(data, hypers)
        parameters = sample_parameters(post_hyperparameters, 1)[0]
    else:
        parameters = None

    return parameters



"""
Tahmid F. Mehdi
Moses Lab, University of Toronto
Probability Functions for Several Exponential Families
February 5, 2018

Copyright 2018 Tahmid Mehdi
This file is part of dphmix.

dphmix is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

dphmix is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with dphmix.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
import math
import numpy as np
import scipy.special as sp


def normal_pdf(x, mu, tau):
    """
    PDF of normal random variable x

    :param float x: data
    :param float mu: mean
    :param float>0 tau:  precision
    :return: probability
    :rtype: 0<=float<=1
    """

    return np.sqrt(tau/(2*np.pi))*np.exp(-(tau*(x-mu)**2)/2)


def bernoulli_pmf(x, p):
    """
    PMF of bernoulli random variable x

    :param int x: binary data. {0,1}
    :param 0<=float<=1 p: probability of success
    :return: probability
    :rtype: 0<=float<=1
    """

    return p**x*(1-p)**(1-x)


def poisson_pmf(x, lam):
    """
    PMF of poisson random variable x

    :param int>=0 x: data
    :param float>=0 lam: rate parameter
    :return: probability
    :rtype: 0<=float<=1
    """

    return lam**x*np.exp(-lam)/math.factorial(x)


def gamma_pdf(x, a, b):
    """
    PDF of gamma random variable x

    :param float>=0 x: data
    :param float>0 a: shape
    :param float>0 b: rate
    :return: probability
    :rtype: 0<=float<=1
    """

    return b**a*x**(a-1)*np.exp(-b*x)/sp.gamma(a)


def beta_pdf(x, a, b):
    """
    PDF of beta random variable x

    :param float x: data. [0,1]
    :param float>0 a: shape
    :param float>0 b: shape
    :return: probability
    :rtype: 0<=float<=1
    """

    return (x**(a-1)*(1-x)**(b-1))/sp.beta(a, b)


def ln_normal(x, ln_tau, mu, tau):
    """
    ln probability of Normal random variable x

    :param float x: data
    :param float ln_tau: ln(precision)
    :param float mu: mean
    :param float>0 tau: precision
    :return: lnP(x)
    :rtype: float<=0
    """

    return 0.5*(ln_tau-np.log(2*np.pi)-tau*(x-mu)**2)


def ln_bernoulli(x, ln_p, ln_ip):
    """
    ln probability of Bernoulli random variable x

    :param int x: data. {0,1}
    :param float<=0 ln_p: ln(p)
    :param float<=0 ln_ip: ln(1-p)
    :return: lnP(x)
    :rtype: float<=0
    """

    return x*ln_p+(1-x)*ln_ip


def ln_poisson(x, ln_lam, lam):
    """
    ln probability of Poisson random variable x

    :param int>=0 x: data
    :param float ln_lam: ln(rate)
    :param float>=0 lam: rate parameter
    :return: lnP(x)
    :rtype: float<=0
    """

    return x*ln_lam-math.log(math.factorial(x))-lam


def ln_normalgamma(x, y, ln_y, nu, rho, a, b):
    """
    ln probability of NormalGamma random variable (x,y)

    :param float x: data
    :param float>0 y: data
    :param float ln_y: ln(y)
    :param float nu: mean
    :param float>0 rho: precision
    :param float>0 a: shape
    :param float>0 b: rate
    :return: lnP(x,y)
    :rtype: float<=0
    """

    return 0.5*(math.log(rho/(2*np.pi))+ln_y-rho*y*(x-nu)**2)+a*math.log(b)-sp.gammaln(a)+(a-1)*ln_y-b*y


def ln_beta(ln_x, ln_ix, a, b):
    """
    ln probability of Beta random variable

    :param float<=0 ln_x: ln(x). x is the data
    :param float<=0 ln_ix: ln(1-x)
    :param float>0 a: shape
    :param float>0 b: shape
    :return: lnP(x)
    :rtype: float<=0
    """

    return (a-1)*ln_x+(b-1)*ln_ix-sp.betaln(a, b)


def ln_gamma(x, ln_x, a, b):
    """
    ln probability of Gamma random variable x

    :param float>=0 x: data
    :param float ln_x: ln(x)
    :param float>0 a: shape
    :param float>0 b: rate
    :return: lnP(x)
    :rtype: float<=0
    """

    return a*math.log(b)-sp.gammaln(a)+(a-1)*ln_x-b*x



"""
Tahmid F. Mehdi
Moses Lab, University of Toronto
Unsupervised and Constrained Dirichlet Heterogeneous Mixture of Exponential Families with Gibbs Sampling
July 19, 2018

References:
RM Neal. (2000). Markov Chain Sampling Methods for Dirichlet Process Mixture Models. Journal of Computational and
Graphical Statistics, 9(2), 249-265.
A Vlachos, A Korhonen & Z Ghahramani. (2009). Unsupervised and Constrained Dirichlet Process Mixture Models for Verb
Clustering. Proceedings of the EACL 2009 Workshop on GEMS: GEometical Models of Natural Language Semantics, 74-82.

Copyright 2018 Tahmid Mehdi
This file is part of dphmix.

dphmix is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

dphmix is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with dphmix.  If not, see <http://www.gnu.org/licenses/>.
"""

from __future__ import division
from .gibbsHelpers import *
from .clusters import *
from collections import Counter
from itertools import count, filterfalse
from joblib import Parallel, delayed
from .utils import *


class GibbsDPHM:
    """
Implements Algorithm 2 from Neal (2000) with conjugate priors for Dirichlet Process Mixture Models where the independent
variates follow Normal, Bernoulli & Poisson distributions. It can handle must-link & can't-link constraints

If x has n_cts, n_bin & n_ord continuous, binary & ordinal variates, respectively, then
hyperparameters are contained in dictionaries where keys=['nu','rho','a','b','gamma','delta','eps','zeta'] where
the values of 'nu','rho','a' & 'b' must be lists of numbers with length n_cts
the values of 'gamma' & 'delta' must be lists of numbers with length n_bin
the values of 'eps' & 'zeta' must be lists of numbers with length n_ord

Parameters are contained in dictionaries where keys=['mu','tau','p','lam'] where
the values of 'mu' & 'tau' must be lists of numbers with length n_cts
the value of 'p' must be a list of numbers with length n_bin
the value of 'lam' must be a list of numbers with length n_ord

x | parameters ~ N(mu[1],1/tau[1])..N(mu[n_cts],1/tau[n_cts])Bern(p[1])..Bern(p[n_bin])Poi(lam[1])..Poi(lam[n_ord])
where (mu[i],tau[i]) ~ NormalGamma(nu[i],rho[i],a[i],b[i]), i=1..n_cts
      p[i] ~ Beta(gamma[i],delta[i]), i=1..n_bin
      lam[i] ~ Gamma(eps[i],zeta[i]), i=1..n_ord
    """
    def __init__(self, alpha, iterations, max_clusters, n_jobs=1):
        """
        Constructs a DirichletHeterogeneousMixture object

        :param float>0 alpha: the alpha parameter of the Dirichlet Process distribution
        :param int>=1 iterations: maximum number of iterations
        :param int>=2 max_clusters: maximum number of clusters
        :param int>=1 n_jobs: number of cores to use
        """

        assert alpha > 0, 'alpha must be positive'
        assert iterations >= 1 and isinstance(iterations, int), 'iterations must be a positive integer'
        assert max_clusters >= 2 and isinstance(max_clusters, int), 'max_clusters must be an integer at least 2'
        assert n_jobs >= 1 and isinstance(n_jobs, int), 'n_jobs must be an integer at least 1'
        self.alpha = alpha
        self.iterations = iterations
        self.max_clusters = max_clusters
        self.n_jobs = n_jobs
        self.clusters = None
        self.phi = None

    def get_params(self):
        """
        Returns a dictionary of arguments for the object

        :return: arguments for object
        :rtype: dict
        """
        return dict(alpha=self.alpha, iterations=self.iterations, max_clusters=self.max_clusters, n_jobs=self.n_jobs)

    def __repr__(self):
        return "GibbsDPHM(alpha={}, iterations={}, max_clusters={}, n_jobs={})".\
            format(self.alpha, self.iterations, self.max_clusters, self.n_jobs)

    def __str__(self):
        return "GibbsDPHM(alpha={}, iterations={}, max_clusters={}, n_jobs={})".\
            format(self.alpha, self.iterations, self.max_clusters, self.n_jobs)

    def _gibbs(self, X, hyperparameters, ml=[], cl=[]):
        """
        Clusters X with Gibbs Sampler given must-link constraints ml & can't-link constraints cl

        :param pd.DataFrame X: data
        :param hyperparameters hyperparameters: hyperparameters for distributions
        :param list ml: each element is a list of indices of points that must be in the same cluster
        :param list cl: each element is a list of indices of points that cannot be in the same cluster
        :return: cluster assignments for each observation and cluster parameters
        :rtype: GibbsCluster
        """

        n = len(X)
        ml_flat = [idx for link in ml for idx in link]
        not_ml = [i for i in range(n) if i not in ml_flat]

        # create mapping of data indices to indices of observations they cannot be clustered with
        cantLink = {}
        for link in cl:
            for idx in link:
                if idx in cantLink:
                    cantLink[idx].extend([m for m in link if m != idx])
                else:
                    cantLink[idx] = [m for m in link if m != idx]
        # create mapping of must-links to indices of observations they cannot be clustered with
        cantLink_ml = {}
        integral_ml = [0]*len(ml)
        for link in range(len(ml)):
            cantLink_ml[link] = []
            for idx in ml[link]:
                if idx in cantLink:
                    cantLink_ml[link].extend(cantLink[idx])

            integral_ml[link] = integrate(X.iloc[ml[link]], hyperparameters)

        c = np.array([0]*n)  # randomly assign clusters
        # initialize cluster parameters. If unsupervised then just initialize 1 cluster. If semi-supervised then
        # initialize clusters for each label as well as cluster 0 for unlabeled data
        phi = Parallel(n_jobs=self.n_jobs)(delayed(sample_parameters_posterior)
                                           (data=X.iloc[c == clust], hypers=hyperparameters)
                                           for clust in range(self.max_clusters))
        # calculate integrals
        integral = X.apply(lambda x: integrate(x, hyperparameters), axis=1)
        for state in range(self.iterations):
            print("Running MCMC Step %s" % str(state + 1))
            for i in not_ml:
                c_noi = np.delete(c, i, axis=0)  # clusters without X[i]
                n_noi = Counter(c_noi)  # count frequency of each cluster
                if c[i] not in c_noi:  # X[i] is the only one from its cluster
                    phi[c[i]] = None  # delete its cluster

                uniq_clusters = np.unique(c_noi)
                # remove clusters of data that the ith point cannot cluster with
                if i in cantLink:
                    uniq_clusters = [clust for clust in uniq_clusters if clust not in map(lambda a: c[a], cantLink[i])]
                # dictionary of probabilities of each cluster
                prob_c = {}
                # set probabilities for existing clusters according to Chinese Restaurant Process
                for clust in uniq_clusters:
                    prob_c[clust] = n_noi[clust]*likelihood(X.iloc[i], phi[clust])
                # find the lowest non-negative integer that hasn't indexed a cluster
                new_cluster = next(filterfalse(uniq_clusters.__contains__, count(0)))
                # set probability for new cluster
                prob_c[new_cluster] = self.alpha*integral[i]
                c[i] = assign_cluster(prob_c)

                if c[i] == new_cluster:  # if X[i] was assigned the new cluster
                    assert new_cluster < self.max_clusters, 'Number of clusters exceeded max_clusters'
                    # set parameters for new cluster
                    post_hyperparameters = posterior_hyperparameters(X.iloc[i], hyperparameters)
                    phi[new_cluster] = sample_parameters(post_hyperparameters, 1)[0]
            # pass through data in must-link constraints
            for link in range(len(ml)):
                c_noi = np.delete(c, ml[link], axis=0)  # clusters without X[i]
                n_noi = Counter(c_noi)  # count frequency of each cluster

                uniq_clusters = np.unique(c_noi)
                # remove clusters of points the group cannot cluster with from uniq_clusters
                if i in cantLink_ml:
                    uniq_clusters = [clust for clust in uniq_clusters if clust not in map(lambda a: c[a], cantLink_ml[link])]
                # dictionary of probabilities of each cluster
                prob_c = {}
                # set probabilities for existing clusters according to Chinese Restaurant Process
                for clust in uniq_clusters:
                    # use joint probability of data in ml[link]
                    prob_c[clust] = n_noi[clust]*likelihood(X.iloc[ml[link]], phi[clust])
                # find the lowest non-negative integer that hasn't indexed a cluster
                new_cluster = next(filterfalse(uniq_clusters.__contains__, count(0)))
                # set probability for new cluster
                prob_c[new_cluster] = self.alpha*integral_ml[link]
                c[ml[link]] = assign_cluster(prob_c)

                if c[ml[link][0]] == new_cluster:  # if X[i] was assigned the new cluster
                    assert new_cluster < self.max_clusters, 'Number of clusters exceeded max_clusters'
                    # set parameters for new cluster
                    post_hyperparameters = posterior_hyperparameters(X.iloc[ml[link]], hyperparameters)
                    phi[new_cluster] = sample_parameters(post_hyperparameters, 1)[0]
            # resample parameters for all clusters from the posterior
            phi = Parallel(n_jobs=self.n_jobs)(delayed(sample_parameters_posterior)
                                               (data=X.iloc[c == clust], hypers=hyperparameters)
                                               for clust in range(self.max_clusters))

        # remove Nones from phi
        self.phi = [theta for theta in phi if theta is not None]
        # map cluster indices so they're all consecutive integers
        uniq_c = np.unique(c)
        self.clusters = list(map(lambda clust: np.where(uniq_c == clust)[0][0], c))
        solution = GibbsCluster(self.clusters, self.phi)
        print("Done")
        return solution

    def fit_predict(self, X, ml=[], cl=[], hyperparameters=None):
        """
        Fits a model & predict clusters for each observation (row) in X given must-link constraints ml and can't-link
        constraints cl

        :param pd.DataFrame X: data
        :param list ml: each element is a list of indices of points that must be in the same cluster
        :param list cl: each element is a list of indices of points that cannot be in the same cluster
        :param hyperparameters: hyperparameters of distributions for the variates
        :type hyperparameters: hyperparameters or None
        :return: the clustering solution
        :rtype: GibbsCluster
        """

        assert isinstance(X, pd.DataFrame), 'X must be a pandas.DataFrame'
        ml_flat = [idx for link in ml for idx in link]
        cl_flat = [idx for link in cl for idx in link]
        assert not(is_duplicate(ml_flat)), 'Integers in ml are not unique'
        assert (0 <= np.array(ml_flat)).all() and (np.array(ml_flat) < len(X)).all() \
            and all([isinstance(item, int) for item in ml_flat]), 'Elements of ml must be integers in [0,len(X))'
        assert (0 <= np.array(cl_flat)).all() and (np.array(cl_flat) < len(X)).all() \
            and all([isinstance(item, int) for item in cl_flat]), 'Elements of cl must be integers in [0,len(X))'
        # order the columns of X so Gaussian variates come first, then Bernoullis then Poissons
        X, n_cts, n_bin, n_ord = sort_features(X)

        if hyperparameters is None:
            # set prior hyperparameters
            nu = np.mean(X[X.columns[:n_cts]])
            rho = [1]*n_cts
            a = [1]*n_cts
            b = (np.var(X[X.columns[:n_cts]]))
            gamma = [2]*n_bin
            delta = (2-2*np.mean(X[X.columns[n_cts:(n_cts+n_bin)]]))/np.mean(X[X.columns[n_cts:(n_cts+n_bin)]])
            eps = [1]*n_ord
            zeta = 1/np.mean(X[X.columns[(n_cts+n_bin):]])
            hyperparameters = dict(nu=nu, rho=rho, a=a, b=b, gamma=gamma, delta=delta, eps=eps, zeta=zeta)
        else:  # check lengths of hyperparameters
            assert isinstance(hyperparameters, dict), 'hyperparameters must be a dictionary'
            assert 'nu' in hyperparameters and 'rho' in hyperparameters and 'a' in hyperparameters \
                   and 'b' in hyperparameters and 'gamma' in hyperparameters and 'delta' in hyperparameters \
                   and 'eps' in hyperparameters and 'zeta' in hyperparameters, \
                'hyperparameters needs keys "nu", "rho", "a", "b", "gamma", "delta", "eps" & "zeta"'
            assert len(hyperparameters['nu']) == n_cts and len(hyperparameters['rho']) == n_cts \
                   and len(hyperparameters['a']) == n_cts and len(hyperparameters['b']) == n_cts, \
                'Number of hyperparameters for continuous variates do not match the number of continuous variates'
            assert len(hyperparameters['gamma']) == n_bin and len(hyperparameters['delta']) == n_bin, \
                'Number of hyperparameters for binary variates do not match the number of binary variates'
            assert len(hyperparameters['eps']) == n_ord and len(hyperparameters['zeta']) == n_ord, \
                'Number of hyperparameters for ordinal variates do not match the number of ordinal variates'

        # perform Gibbs sampling
        clust_soln = self._gibbs(X, hyperparameters, ml, cl)

        return clust_soln

    def predict(self, X):
        """
        Predicts clusters for the observations (rows) in X

        :param pd.DataFrame X: data
        :return: the clustering solution
        :rtype: np.array, np.ndarray
        """

        assert self.phi is not None and self.clusters is not None, 'Model has not been fitted'
        # order the columns of X so Gaussian variates come first, then Bernoullis then Poissons
        X, n_cts_data, n_bin_data, n_ord_data = sort_features(X)
        n_cts = len(self.phi[0]['mu'])
        n_bin = len(self.phi[0]['p'])
        n_ord = len(self.phi[0]['lam'])
        assert n_cts+n_bin+n_ord == len(X.columns), 'The features in X do not match the features of the model'
        n = len(X)
        uniq_c = np.unique(self.clusters)  # array of cluster indices
        weights = Counter(self.clusters)  # weights of the clusters
        prob = np.zeros((n, len(uniq_c)))  # initialize probability matrix
        # construct the matrix
        for idx in range(n):
            prob[idx, :] = [weights[clust]*likelihood(X.iloc[idx], self.phi[clust]) for clust in uniq_c]
        # normalize prob
        rowSums = prob.sum(axis=1)[:, None]  # sum probabilities for each observation
        # if all probabilities for an observation are below machine epsilon then it won't be assigned a cluster
        assert 0 not in rowSums, \
            str(list(rowSums).count(0)) + ' observations could not be assigned to a cluster. Increase max_clusters'
        prob = prob / rowSums  # normalize prob
        # pick most probable cluster for each observation
        c = np.array(prob.argmax(axis=1)).reshape(n)
        return c, prob



from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()


setup(name='dphmix',
      version='0.1.0',
      description='Unsupervised and Semi-supervised Dirichlet Process Heterogeneous Mixtures',
      long_description='Implements Dirichlet Process Heterogeneous Mixtures of exponential family distributions for clustering heterogeneous data without choosing the number of clusters. Inference can be performed with Gibbs sampling or coordinate ascent mean-field variational inference. For semi-supervised learning, Gibbs sampling supports must-link and cannot-link constraints. A novel variational inference algorithm was derived to handle must-link constraints.',
      classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
      ],
      url='https://github.com/tahmidmehdi/dphmix',
      download_url='https://github.com/tahmidmehdi/dphmix/archive/v0.1.tar.gz',
      author='Tahmid Mehdi',
      author_email='tfmehdi@cs.toronto.edu',
      license='GNU GPL v3.0',
      packages=['dphmix'],
      install_requires=[
          'numpy',
          'joblib',
          'scipy',
          'pandas',
      ],
      test_suite='nose.collector',
      tests_require=['nose', 'nose-cover3'],
      include_package_data=True,
      zip_safe=False)


